@@ -3,11 +3,14 @@ Datatypes and models for various OpenID messages
 """
 import enum
 import logging
-from typing import Any, List, Literal, Mapping, NewType, Optional, Union
+import time
+from datetime import datetime
+from typing import Any, Callable, List, Literal, Mapping, Optional, Union
 
-from pydantic import Extra, Field, HttpUrl, root_validator
+from pydantic import AnyHttpUrl, Extra, Field, root_validator
 
-from simple_openid_connect.base_data import OpenidBaseModel, OpenidMessage
+from simple_openid_connect.base_data import OpenidBaseModel
+from simple_openid_connect.utils import validate_that
 
 logger = logging.getLogger(__name__)
 
@@ -25,22 +28,22 @@ class ProviderMetadata(OpenidBaseModel):
         extra = Extra.allow
         allow_mutation = False
 
-    issuer: HttpUrl
+    issuer: AnyHttpUrl
     "REQUIRED. URL using the https scheme with no query or fragment component that the OP asserts as its Issuer Identifier This MUST be identical to the iss Claim value in ID Tokens issued from this Issuer."
 
-    authorization_endpoint: HttpUrl
+    authorization_endpoint: AnyHttpUrl
     "REQUIRED. URL of the OP's OAuth 2.0 Authorization Endpoint."
 
-    token_endpoint: Optional[HttpUrl]
+    token_endpoint: Optional[AnyHttpUrl]
     "URL of the OP's OAuth 2.0 Token Endpoint. This is REQUIRED unless only the Implicit Flow is used."
 
-    userinfo_endpoint: Optional[HttpUrl]
+    userinfo_endpoint: Optional[AnyHttpUrl]
     "RECOMMENDED. URL of the OP's UserInfo Endpoint. This URL MUST use the https scheme and MAY contain port, path, and query parameter components."
 
-    jwks_uri: HttpUrl
+    jwks_uri: AnyHttpUrl
     "REQUIRED. URL of the OP's JSON Web Key Set document This contains the signing key(s) the RP uses to validate signatures from the OP The JWK Set MAY also contain the Server's encryption key(s), which are used by RPs to encrypt requests to the Server When both signing and encryption keys are made available, a use (Key Use) parameter value is REQUIRED for all keys in the referenced JWK Set to indicate each key's intended usage Although some algorithms allow the same key to be used for both signatures and encryption, doing so is NOT RECOMMENDED, as it is less secure The JWK x5c parameter MAY be used to provide X.509 representations of keys provided When used, the bare key values MUST still be present and MUST match those in the certificate. "
 
-    registration_endpoint: Optional[HttpUrl]
+    registration_endpoint: Optional[AnyHttpUrl]
     "RECOMMENDED. URL of the OP's Dynamic Client Registration Endpoint"
 
     scopes_supported: Optional[List[str]]
@@ -88,7 +91,9 @@ class ProviderMetadata(OpenidBaseModel):
     request_object_encryption_enc_values_supported: Optional[List[str]]
     "OPTIONAL. JSON array containing a list of the JWE encryption algorithms (enc values) supported by the OP for Request Objects These algorithms are used both when the Request Object is passed by value and when it is passed by reference."
 
-    token_endpoint_auth_methods_supported: Optional[List[str]] = ["client_secret_basic"]
+    token_endpoint_auth_methods_supported: List[str] = Field(
+        default=["client_secret_basic"]
+    )
     "OPTIONAL. JSON array containing a list of Client Authentication methods supported by this Token Endpoint The options are client_secret_post, client_secret_basic, client_secret_jwt, and private_key_jwt, as described in Section 9 of OpenID Connect Core 1.0 Other authentication methods MAY be defined by extensions. If omitted, the default is client_secret_basic -- the HTTP Basic Authentication Scheme specified in Section 2.3.1 of OAuth 2.0 [RFC6749]."
 
     token_endpoint_auth_signing_alg_values_supported: Optional[List[str]]
@@ -103,7 +108,7 @@ class ProviderMetadata(OpenidBaseModel):
     claims_supported: Optional[List[str]]
     "RECOMMENDED. JSON array containing a list of the Claim Names of the Claims that the OpenID Provider MAY be able to supply values for Note that for privacy or other reasons, this might not be an exhaustive list."
 
-    service_documentation: Optional[HttpUrl]
+    service_documentation: Optional[AnyHttpUrl]
     "OPTIONAL. URL of a page containing human-readable information that developers might want or need to know when using the OpenID Provider In particular, if the OpenID Provider does not support Dynamic Client Registration, then information on how to register Clients needs to be provided in this documentation."
 
     claims_locales_supported: Optional[List[str]]
@@ -124,13 +129,13 @@ class ProviderMetadata(OpenidBaseModel):
     require_request_uri_registration: Optional[bool] = False
     "OPTIONAL. Boolean value specifying whether the OP requires any request_uri values used to be pre-registered using the request_uris registration parameter Pre-registration is REQUIRED when the value is true. If omitted, the default value is false."
 
-    op_policy_uri: Optional[HttpUrl]
+    op_policy_uri: Optional[AnyHttpUrl]
     "OPTIONAL. URL that the OpenID Provider provides to the person registering the Client to read about the OP's requirements on how the Relying Party can use the data provided by the OP The registration process SHOULD display this URL to the person registering the Client if it is given."
 
-    op_tos_uri: Optional[HttpUrl]
+    op_tos_uri: Optional[AnyHttpUrl]
     "OPTIONAL. URL that the OpenID Provider provides to the person registering the Client to read about OpenID Provider's terms of service The registration process SHOULD display this URL to the person registering the Client if it is given. "
 
-    end_session_endpoint: Optional[HttpUrl]
+    end_session_endpoint: Optional[AnyHttpUrl]
     "REQUIRED, if supported by OP. URL at the OP to which an RP can perform a redirect to request that the End-User be logged out at the OP."
 
     frontchannel_logout_supported: bool = Field(default=False)
@@ -145,7 +150,7 @@ class ProviderMetadata(OpenidBaseModel):
     backchannel_logout_session_supported: bool = Field(default=False)
     "OPTIONAL. Boolean value specifying whether the OP can pass a sid (session ID) Claim in the Logout Token to identify the RP session with the OP. If supported, the sid Claim is also included in ID Tokens issued by the OP. If omitted, the default value is false. "
 
-    introspection_endpoint: Optional[HttpUrl]
+    introspection_endpoint: Optional[AnyHttpUrl]
     "OPTIONAL. URL that the OpenID Provider provides to resource servers to introspect access tokens in accordance to `RFC7662: OAuth 2.0 Token Introspection <https://www.rfc-editor.org/rfc/rfc7662>`_."
 
 
@@ -163,7 +168,7 @@ class IdToken(OpenidBaseModel):
         extra = Extra.allow
         allow_mutation = False
 
-    iss: HttpUrl
+    iss: AnyHttpUrl
     "REQUIRED. Issuer Identifier for the Issuer of the response The iss value is a case sensitive URL using the https scheme that contains scheme, host, and optionally, port number and path components and no query or fragment components."
 
     sub: str
@@ -196,8 +201,101 @@ class IdToken(OpenidBaseModel):
     sid: Optional[str]
     "OPTIONAL. Session ID - String identifier for a Session. This represents a Session of a User Agent or device for a logged-in End-User at an RP. Different sid values are used to identify distinct sessions at an OP. The sid value need only be unique in the context of a particular issuer. Its contents are opaque to the RP."
 
+    def validate_extern(
+        self,
+        issuer: str,
+        client_id: str,
+        nonce: Union[str, None] = None,
+        extra_trusted_audiences: List[str] = [],
+        min_iat: float = 0,
+        validate_acr: Union[Callable[[str], None], None] = None,
+        min_auth_time: float = 0,
+    ) -> None:
+        """
+        Validate this ID-Token with external data for consistency
 
-class UserinfoRequest(OpenidMessage):
+        :param issuer: The issuer that this token is supposed to originate from.
+            Should usually be :data:`ProviderMetadata.issuer`
+        :param client_id: The client id of this client
+        :param nonce: The nonce that was used during authentication.
+            It is carried over by the OP into ID-Tokens and must now match.
+        :param extra_trusted_audiences: Which token audiences (client ids) to consider trusted beside this client's own client_id.
+            This is usually an empty list but if the token is intended to be used by more than one client, all of these need to be listed in the tokens :data:`IdToken.aud` field, and they all need to be known and trusted by this client.
+        :param min_iat: Minimum value that the tokens :data:`IdToken.iat` claim must be.
+            This value is a posix timestamp and defaults to 0 which allows arbitrarily old `iat` dates.
+        :param validate_acr: A callable that receives this tokens :data:`IdToken.acr` value and must perform it own validation.
+            This is necessary because the value of acr is outside OpenId-Connect specification and usage specific.
+            If not given, acr is assumed to always be valid.
+        :param min_auth_time: The point in time which is considered the minimum at which a user should have authenticated.
+            It basically means that if the user was authenticated very far in the past and reused their session, the time at which the original authentication took place must be greater than this value.
+            This is only validated if the :data:`IdToken.auth_time` is present in the token.
+            This value is a posix timestamp and default to 0 which allows arbitrarily old `auth_time` dates.
+
+        :raises ValidationError: if the validation fails
+        """
+        # this method implements https://openid.net/specs/openid-connect-core-1_0.html#IDTokenValidation
+
+        # 2. validate issuer
+        validate_that(self.iss == issuer, "ID-Token was issued from unexpected issuer")
+
+        # 3. validate audience
+        if isinstance(self.aud, str):
+            validate_that(
+                self.aud == client_id,
+                "ID-Token's audience does not contain own client_id",
+            )
+        elif isinstance(self.aud, list):
+            validate_that(
+                client_id in self.aud,
+                "ID-Token's audience does not contain own client_id",
+            )
+            validate_that(
+                all(i in extra_trusted_audiences for i in self.aud),
+                "Not all of the ID-Token's audience are trusted",
+            )
+
+        # 4. validate that an azp claim is present if required
+        if isinstance(self.aud, list) and len(self.aud) > 1:
+            validate_that(
+                self.azp is not None,
+                "ID-Token does not contain azp claim but is required to because it is issued to more than 1 audience",
+            )
+
+        # 5. validate azp claim value
+        if self.azp is not None:
+            validate_that(
+                self.azp == client_id,
+                "The ID-Token was not issued to this client (azp claim mismatch)",
+            )
+
+        # 9. validate expiry
+        validate_that(self.exp > time.time(), "The ID-Token is expired")
+
+        # 10. validate iat
+        validate_that(
+            self.iat >= min_iat, "The ID-Token was issued too far in the past"
+        )
+
+        # 11. validate nonce
+        if self.nonce is not None:
+            validate_that(
+                self.nonce == nonce,
+                "The ID-Token's nonce does not match its expected value",
+            )
+
+        # 12. validate acr
+        if self.acr is not None and validate_acr is not None:
+            validate_acr(self.acr)
+
+        # 13. validate auth_time
+        if self.auth_time is not None:
+            validate_that(
+                self.auth_time >= min_auth_time,
+                "The session associated with this ID-Token was authenticated too far in the past",
+            )
+
+
+class UserinfoRequest(OpenidBaseModel):
     """
     A request that can be sent to an OP to request information about a user
     """
@@ -205,7 +303,7 @@ class UserinfoRequest(OpenidMessage):
     pass
 
 
-class UserinfoSuccessResponse(OpenidMessage):
+class UserinfoSuccessResponse(OpenidBaseModel):
     """
     A successful response to a userinfo request containing data about the requested user.
 
@@ -221,7 +319,7 @@ class UserinfoSuccessResponse(OpenidMessage):
     "Subject of this response, basically a unique user id."
 
 
-class UserinfoErrorResponse(OpenidMessage):
+class UserinfoErrorResponse(OpenidBaseModel):
     """
     An error response that is sent back from an OP after requesting user information
     """
@@ -234,7 +332,7 @@ class UserinfoErrorResponse(OpenidMessage):
     error_description: Optional[str]
 
 
-class AuthenticationRequest(OpenidMessage):
+class AuthenticationRequest(OpenidBaseModel):
     """
     An Authentication Request requests that the End-User be authenticated by the Authorization Server.
     """
@@ -285,7 +383,7 @@ class AuthenticationRequest(OpenidMessage):
     "OPTIONAL. Requested Authentication Context Class Reference values Space-separated string that specifies the acr values that the Authorization Server is being requested to use for processing this Authentication Request, with the values appearing in order of preference The Authentication Context Class satisfied by the authentication performed is returned as the acr Claim Value, as specified in Section 2 The acr Claim is requested as a Voluntary Claim by this parameter."
 
 
-class AuthenticationSuccessResponse(OpenidMessage):
+class AuthenticationSuccessResponse(OpenidBaseModel):
     """
     A response that is sent by the Authorization Server if a previous :class:`.AuthenticationRequest` could successfully
     be parsed and handled.
@@ -303,7 +401,7 @@ class AuthenticationSuccessResponse(OpenidMessage):
     "REQUIRED if the `state` parameter was present in the client authorization request The exact value received from the client."
 
 
-class AuthenticationErrorResponse(OpenidMessage):
+class AuthenticationErrorResponse(OpenidBaseModel):
     """
     A response that is sent by the Authorization Server if a previous :class:`AuthenticationRequest` could not be
     understood or handled.
@@ -384,7 +482,7 @@ class AuthenticationErrorResponse(OpenidMessage):
     "REQUIRED if a `state` parameter was present in the client authorization request. The exact value received from the client."
 
 
-class TokenRequest(OpenidMessage):
+class TokenRequest(OpenidBaseModel):
     """
     A Client makes a Token Request by presenting its Authorization Grant (in the form of an Authorization Code) to the Token Endpoint.
     If the Client is a Confidential Client, then it MUST authenticate to the Token Endpoint using the authentication method registered for its client_id.
@@ -411,7 +509,9 @@ class TokenRequest(OpenidMessage):
     "REQUIRED. The refresh token issued to the client."
 
     @root_validator(skip_on_failure=True)
-    def _validate_required_based_on_grant_type(cls, values: Mapping[str, Any]):
+    def _validate_required_based_on_grant_type(
+        cls, values: Mapping[str, Any]
+    ) -> Mapping[str, Any]:
         if values["grant_type"] == "authorization_code":
             assert (
                 values["code"] is not None
@@ -427,14 +527,11 @@ class TokenRequest(OpenidMessage):
         return values
 
 
-class TokenSuccessResponse(OpenidMessage):
+class TokenSuccessResponse(OpenidBaseModel):
     """
     After receiving and validating a valid and authorized :class:`TokenRequest <TokenRequest>` from the Client, the Authorization Server returns a successful response that includes an ID Token and an Access Token
     """
 
-    # TODO Implement ID Token validation
-    # https://openid.net/specs/openid-connect-core-1_0.html#IDTokenValidation
-
     class Config:
         allow_mutation = False
 
@@ -450,6 +547,9 @@ class TokenSuccessResponse(OpenidMessage):
     refresh_token: Optional[str]
     "OPTIONAL. The refresh token, which can be used to obtain new access tokens using the same authorization grant as described in `Section 6 of RFC6749 <https://www.rfc-editor.org/rfc/rfc6749#section-6>`_."
 
+    refresh_expires_in: Optional[int]
+    "OPTIONAL. The lifetime in seconds of the refresh token."
+
     scope: Optional[str]
     "OPTIONAL, if identical to the scope requested by the client; otherwise, REQUIRED. The scope of the access token."
 
@@ -457,7 +557,7 @@ class TokenSuccessResponse(OpenidMessage):
     "ID Token value associated with the authenticated session."
 
 
-class TokenErrorResponse(OpenidMessage):
+class TokenErrorResponse(OpenidBaseModel):
     """
     A response that is sent by the Authorization Server if a previous :class:`.TokenRequest` could not be
     understood or handled.
@@ -502,7 +602,7 @@ class TokenErrorResponse(OpenidMessage):
     "OPTIONAL.  A URI identifying a human-readable web page with information about the error, used to provide the client developer with additional information about the error."
 
 
-class RpInitiatedLogoutRequest(OpenidMessage):
+class RpInitiatedLogoutRequest(OpenidBaseModel):
     """
     Request which a Relying-Party sends to the OP to initiate a user logout
     """
@@ -526,7 +626,7 @@ class RpInitiatedLogoutRequest(OpenidMessage):
     'OPTIONAL. End-User\'s preferred languages and scripts for the user interface, represented as a space-separated list of BCP47 [RFC5646] language tag values, ordered by preference. For instance, the value "fr-CA fr en" represents a preference for French as spoken in Canada, then French (without a region designation), followed by English (without a region designation). An error SHOULD NOT result if some or all of the requested locales are not supported by the OpenID Provider. '
 
 
-class FrontChannelLogoutNotification(OpenidMessage):
+class FrontChannelLogoutNotification(OpenidBaseModel):
     """
     A notification which the Relying-Party receives when a user logs out.
 
@@ -544,7 +644,7 @@ class FrontChannelLogoutNotification(OpenidMessage):
     "Identifier for the Session."
 
 
-class BackChannelLogoutNotification(OpenidMessage):
+class BackChannelLogoutNotification(OpenidBaseModel):
     """
     A notification which the Relying-Party receives from the OP when a user logs out.
 
@@ -555,7 +655,7 @@ class BackChannelLogoutNotification(OpenidMessage):
     "A signed JWT containing a :class:`BackChannelLogoutToken`"
 
 
-class BackChannelLogoutToken(OpenidMessage):
+class BackChannelLogoutToken(OpenidBaseModel):
     """
     OPs send a JWT similar to an ID Token to RPs called a Logout Token to request that they log out.
     This token is sent as part of a :class:`BackChannelLogoutNotification`.
@@ -571,7 +671,7 @@ class BackChannelLogoutToken(OpenidMessage):
         extra = Extra.forbid
 
     class Events(OpenidBaseModel):
-        x: dict = Field(
+        x: Mapping[str, Any] = Field(
             alias="http://schemas.openid.net/event/backchannel-logout",
             default={},
             const=True,
@@ -586,7 +686,7 @@ class BackChannelLogoutToken(OpenidMessage):
     aud: str
     "REQUIRED. Audience(s)"
 
-    iat: str
+    iat: int
     "REQUIRED. Issued at time"
 
     jti: str
@@ -598,8 +698,88 @@ class BackChannelLogoutToken(OpenidMessage):
     sid: Optional[str]
     "OPTIONAL. Session ID - String identifier for a Session. This represents a Session of a User Agent or device for a logged-in End-User at an RP. Different sid values are used to identify distinct sessions at an OP. The sid value need only be unique in the context of a particular issuer. Its contents are opaque to the RP."
 
+    def validate_extern(
+        self,
+        issuer: str,
+        client_id: str,
+        extra_trusted_audiences: List[str] = [],
+        min_iat: float = 0,
+        validate_unique_jti: Union[Callable[[str], None], None] = None,
+        validate_iss_has_sessions: Union[Callable[[str], None], None] = None,
+        validate_sub_has_sessions: Union[Callable[[str], None], None] = None,
+        validate_sid_exists: Union[Callable[[str], None], None] = None,
+    ) -> None:
+        """
+        Validate this ID-Token with external data for consistency
 
-class TokenIntrospectionRequest(OpenidMessage):
+        :param issuer: The issuer that this token is supposed to originate from.
+            Should usually be :data:`ProviderMetadata.issuer`
+        :param client_id: The client id of this client
+        :param extra_trusted_audiences: Which token audiences (client ids) to consider trusted beside this client's own client_id.
+            This is usually an empty list but if the token is intended to be used by more than one client, all of these need to be listed in the tokens :data:`IdToken.aud` field, and they all need to be known and trusted by this client.
+        :param min_iat: Minimum value that the tokens :data:`IdToken.iat` claim must be.
+            This value is a posix timestamp and defaults to 0 which allows arbitrarily old `iat` dates.
+        :param validate_unique_jti: A callable which verifies that the given :data:`BackChannelLogoutToken.jti` value has not been previsouly used.
+            If this parameter is not given, validation is skipped.
+        :param validate_iss_has_sessions: A callable which verifies that the logout tokens :data:`iss <BackChannelLogoutToken.iss>` (issuer) has open sessions on this app.
+            If this parameter is not given, validation is skipped.
+        :param validate_sub_has_sessions: A callable which verifies that the logout tokens :data:`sub <BackChannelLogoutToken.sub>` (subject) has one or more open session on this app.
+            If this parameter is not given or the token contains no `sub` claim, validation is skipped.
+        :param validate_sid_exists:  A callable which verifies that the logout tokens :data:`sid <BackChannelLogoutToken.sid>` (session id) is a currently open session on this app.
+            If this parameter is not given or the token contains no `sid` claim, validation is skipped.
+
+        :raises ValidationError: if the validation fails
+        """
+        # this method implements https://openid.net/specs/openid-connect-backchannel-1_0.html#Validation
+
+        # 4. validate iss
+        validate_that(self.iss == issuer, "ID-Token was issued from unexpected issuer")
+
+        # 4. validate audience
+        if isinstance(self.aud, str):
+            validate_that(
+                self.aud == client_id,
+                "ID-Token's audience does not contain own client_id",
+            )
+        elif isinstance(self.aud, list):
+            validate_that(
+                client_id in self.aud,
+                "ID-Token's audience does not contain own client_id",
+            )
+            validate_that(
+                all(i in extra_trusted_audiences for i in self.aud),
+                "Not all of the ID-Token's audience are trusted",
+            )
+
+        # 4. validate iat
+        validate_that(
+            self.iat >= min_iat, "The ID-Token was issued too far in the past"
+        )
+
+        # 5. validate that one of sub or sid (or both) is present
+        validate_that(
+            not (self.sub is None and self.sid is None),
+            "Neither sub nor sid claim is present in Backchannel-Logout-Token",
+        )
+
+        # 8. optionally validate jti
+        if validate_unique_jti is not None:
+            validate_unique_jti(self.jti)
+
+        # 9.optionally validate iss
+        if validate_iss_has_sessions is not None:
+            validate_iss_has_sessions(self.iss)
+
+        # 10. optionally validate sub
+        if validate_sub_has_sessions is not None and self.sub is not None:
+            validate_sub_has_sessions(self.sub)
+
+        # 11. optionally validate sid
+        if validate_sid_exists is not None and self.sid is not None:
+            validate_sid_exists(self.sid)
+
+
+class TokenIntrospectionRequest(OpenidBaseModel):
     """
     The protected resource server calls the introspection endpoint using an HTTP POST request with this request formatted as `application/x-www-form-urlencoded` data.
 
@@ -621,7 +801,7 @@ class TokenIntrospectionRequest(OpenidMessage):
     'OPTIONAL. A hint about the type of the token submitted for introspection. The protected resource MAY pass this parameter to help the authorization server optimize the token lookup. If the server is unable to locate the token using the given hint, it MUST extend its search across all of its supported token types. An OP MAY ignore this parameter, particularly if it is able to detect the token type automatically. Values for this field are defined in the "OAuth Token Type Hints" registry defined in OAuth Token Revocation `RFC7009: OAuth 2.0 Token Revocation <https://www.rfc-editor.org/rfc/rfc7009>`_.'
 
 
-class TokenIntrospectionSuccessResponse(OpenidMessage):
+class TokenIntrospectionSuccessResponse(OpenidBaseModel):
     """
     A message with which an OP responds to :class:`TokenIntrospectionRequest`\s and which contains information about the provided token.
 
